from allianceutils.util import camelize
from rest_framework.renderers import JSONRenderer


class CamelCaseJSONRenderer(JSONRenderer):
    """Renderer that recursively turns underscore-cased keys into camel-cased keys

    This can be set globally on the [DEFAULT_RENDERER_CLASSES](https://www.django-rest-framework.org/api-guide/settings/#default_renderer_classes)
    setting or on a ViewSet on the `renderer_classes` property.
    """

    def render(self, data, *args, **kwargs):
        data = camelize(data)
        return super().render(data, *args, **kwargs)



from typing import Collection


class SerializerOptInFieldsMixin:
    """
    Regulates fields exposed by default & as requested based on query parameters or context.

    Pass 'include_fields' / 'opt_in_fields' thru query params or context to use.
    multiple fields can either be separated by comma eg,
    /?include_fields=first_name,email&opt_in_fields=gait_recognition_prediction
    or passed in the traditional list fashion eg,
    /?include_fields=first_name&include_fields=email&opt_in_fields=gait_recognition_prediction
    or mixed eg,
    /?include_fields=first_name,email&include_fields=boo

    1. By default, all "fields" defined in serializer, minus those listed in "opt_in_fields" would be returned.
    2. If "include_fields" is supplied, only fields requested this way would be returned.
    3. If "opt_in_fields" is supplied, fields requested this way PLUS fields from #1 or #2 would be returned.

    Pinned fields are always returned. (currently pk only if pk exists)
    """

    def get_pinned_fields(self) -> Collection[str]:
        """
        Get by-default pinned fields. Pinned fields are fields always returned regardless of include_fields inclusions.
        Override on serializer to customize.

        Currently, the only pinned field is pk of model (with ModelSerializer).

        :return: [] - a list of pinned fields.
        """

        try:
            return [self.Meta.model._meta.pk.name]
        except AttributeError:
            pass
        return []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not hasattr(self, "context"):
            # serializer invoked without context - inspection?
            return

        pinned_fields = self.get_pinned_fields()

        fields_to_exclude = getattr(self.Meta, "opt_in_only_fields", [])

        fields_to_include = (
            (
                "request" in self.context
                and "include_fields" in self.context["request"].query_params
                and self.context["request"].query_params["include_fields"]
            )
            or ("include_fields" in self.context and self.context["include_fields"])
            or set(self.Meta.fields).difference(set(fields_to_exclude))
        )

        if isinstance(fields_to_include, str):
            fields_to_include = [fields_to_include]

        fields_to_include_list = []
        for f in fields_to_include:
            if f.find(",") != -1:
                fields_to_include_list += f.split(",")
            else:
                fields_to_include_list.append(f)

        opt_in_fields_to_include = (
            (
                "request" in self.context
                and "opt_in_fields" in self.context["request"].query_params
                and self.context["request"].query_params["opt_in_fields"]
            )
            or ("opt_in_fields" in self.context and self.context["opt_in_fields"])
            or []
        )

        for f in list(self.fields.keys()):
            if (
                f not in pinned_fields
                and f not in fields_to_include_list
                and f not in opt_in_fields_to_include
            ):
                del self.fields[f]

        return



from distutils.util import strtobool
import json

from allianceutils.util.camel_case import underscoreize
from django.conf import settings
from rest_framework.exceptions import ParseError
from rest_framework.parsers import JSONParser
from rest_framework.parsers import MultiPartParser


class CamelCaseJSONParser(JSONParser):
    """Parser that recursively turns camelcase keys into underscored keys for JSON data

    This can be set globally on the [DEFAULT_PARSER_CLASSES](https://www.django-rest-framework.org/api-guide/settings/#default_parser_classes)
    setting or on a ViewSet on the `parser_classes` property.
    """

    def underscoreize(self, data, **kwargs):
        """Recursively turn camelcase keys into underscored keys"""
        return underscoreize(data, **kwargs)

    def parse(self, stream, media_type=None, parser_context=None):
        parser_context = parser_context or {}
        encoding = parser_context.get("encoding", settings.DEFAULT_CHARSET)

        try:
            data = stream.read().decode(encoding)
            return self.underscoreize(json.loads(data))
        except ValueError as exc:
            raise ParseError(f"JSON parse error") from exc


class CamelCaseMultiPartJSONParser(MultiPartParser):
    """Parser that recursively turns camelcase keys into underscored keys for JSON data and handles file uploads
    
    This parser supports receiving JSON data where a field value anywhere in the structure can be a file.

    This is achieved on the frontend by converting a structure like:

    ```js
    {
        name: 'Test',
        photo: File,
    }
    ```

    And converting it to

    ```js
    {
        name: 'Test',
        photo: '____ATTACHED_FILE_ID_1',
    }
    ```

    This is then set on a field `jsonData` and the file is set on `____ATTACHED_FILE_ID_1` and submitted
    as multipart.

    This parser then handles parsing the JSON data into a dict and setting each attached file on the
    correct key in the dict.

    Note that this works with nested data (ie. any File anywhere in a nested JSON structure is supported).

    To activate this behaviour the `X-MultiPart-JSON` header must be set to '1' or 'true'. If this header
    is not set it falls back to the default behaviour of MultiPartParser

    This can be set globally on the [DEFAULT_PARSER_CLASSES](https://www.django-rest-framework.org/api-guide/settings/#default_parser_classes)
    setting or on a ViewSet on the `parser_classes` property.

    Example frontend code to activate:

    ```js
    let fileCount = 0;
    const files = {};
    const replacer = (key, value) => {
        if (value instanceof File) {
            const id = `____ATTACHED_FILE_ID_${fileCount++}`;
            files[id] = value;
            return id;
        }
        return value;
    };
    const stringifiedData = JSON.stringify(data, replacer);
    const body = new FormData();
    const body.append('jsonData', stringifiedData);
    for (const [fileKey, file] of Object.entries(files)) {
        body.append(fileKey, file);
    }
    // eg. using a presto Endpoint
    await myEndpoint.execute({
        body,
        headers: {
            // Remove default content type from endpoint (eg. json)
            'Content-Type': undefined,
            'X-MultiPart-JSON': true,
        },
    });
    ```
    """

    def underscoreize(self, data, **kwargs):
        """Recursively turn camelcase keys into underscored keys"""
        return underscoreize(data, **kwargs)

    def parse(self, stream, media_type=None, parser_context=None):
        data_and_files = super().parse(stream, media_type, parser_context)
        request = parser_context["request"]
        if strtobool(request.META.get("HTTP_X_MULTIPART_JSON", "0")):

            def hook(value):
                transformed_values = []
                for (k, v) in value:
                    if type(v) == str and v.startswith("____ATTACHED_FILE_ID_"):
                        final_v = data_and_files.files.get(v)
                    else:
                        final_v = v
                    transformed_values.append((k, final_v))
                return dict(transformed_values)

            json_data = json.loads(
                data_and_files.data.get("jsonData"), object_pairs_hook=hook
            )
            return self.underscoreize(json_data)
        return super().parse(stream, media_type, parser_context)



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="presto_drf",
    version="0.0.2",
    author="Alliance Software",
    author_email="support@alliancesoftware.com.au",
    packages=["presto_drf"],  # this must be the same as the name above
    include_package_data=True,
    description="Integration package for using django-rest-framework with prestojs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["djangorestframework>=3.8"],
    url="https://github.com/prestojs/prestojs/",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=["rad", "rapid application development", "drf", "rest", "django"],
    python_requires=">=3.6",
)
